/*
 * Copyright (c) 1999-2000 Image Power, Inc. and the University of
 *   British Columbia.
 * Copyright (c) 2001 Michael David Adams.
 * All rights reserved.
 *
 * The complete licensing terms for this software can be found in the file
 * named "LICENSE" in this software distribution.
 */

/*
 * JasPer Transcoder Program
 *
 * $Id$
 */

/******************************************************************************\
* Includes.
\******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <time.h>

#include "stream.h"
#include "image.h"
#include "flt.h"
#include "getoptx.h"
#include "image.h"
#include "memalloc.h"
#include "util.h"
#include "version.h"

/******************************************************************************\
*
\******************************************************************************/

#define OPTSMAX	4096

/******************************************************************************\
* Types.
\******************************************************************************/

/* Encoder command line options. */

typedef struct {

	char *infile;
	/* The input image file. */

	int infmt;
	/* The input image file format. */

	char *inopts;
	char inoptsbuf[OPTSMAX + 1];

	char *outfile;
	/* The output image file. */

	int outfmt;

	char *outopts;
	char outoptsbuf[OPTSMAX + 1];

	int verbose;
	/* Verbose mode. */

	int version;

	int_fast32_t cmptno;

} cmdopts_t;

/******************************************************************************\
* Local prototypes.
\******************************************************************************/

void image_initfmts();
cmdopts_t *cmdopts_parse(int argc, char **argv);
void cmdopts_destroy(cmdopts_t *cmdopts);
void cmdusage();
void cmdinfo();
int addopt(char *optstr, int maxlen, char *s);

/******************************************************************************\
* Global data.
\******************************************************************************/

char *cmdname = "";

/******************************************************************************\
* Code.
\******************************************************************************/

int main(int argc, char **argv)
{
	image_t *image;
	cmdopts_t *cmdopts;
	stream_t *in;
	stream_t *out;
	clock_t startclk;
	clock_t endclk;
	long dectime;
	long enctime;
	int_fast16_t numcmpts;
	int i;

	/* Determine the base name of this command. */
	if ((cmdname = strrchr(argv[0], '/'))) {
		++cmdname;
	} else {
		cmdname = argv[0];
	}

	image_initfmts();

	/* Parse the command line options. */
	if (!(cmdopts = cmdopts_parse(argc, argv))) {
		fprintf(stderr, "cannot parse command line\n");
		exit(EXIT_FAILURE);
	}

	if (cmdopts->verbose) {
		cmdinfo();
	}

	if (cmdopts->version) {
		printf("%s\n", VERSION);
		exit(EXIT_SUCCESS);
	}

	/* Open the input image file. */
	if (cmdopts->infile) {
		/* The input image is to be read from a file. */
		if (!(in = stream_fopen(cmdopts->infile, "rb"))) {
			fprintf(stderr, "cannot open input image file %s\n",
			  cmdopts->infile);
			exit(EXIT_FAILURE);
		}
	} else {
		/* The input image is to be read from standard input. */
		if (!(in = stream_fdopen(0, "rb"))) {
			fprintf(stderr, "cannot open standard input\n");
			exit(EXIT_FAILURE);
		}
	}

	/* Open the output image file. */
	if (cmdopts->outfile) {
		/* The output image is to be written to a file. */
		if (!(out = stream_fopen(cmdopts->outfile, "w+b"))) {
			fprintf(stderr, "cannot open output image file %s\n",
			  cmdopts->outfile);
			exit(EXIT_FAILURE);
		}
	} else {
		/* The output image is to be written to standard output. */
		if (!(out = stream_fdopen(1, "w+b"))) {
			fprintf(stderr, "cannot open standard output\n");
			exit(EXIT_FAILURE);
		}
	}

	/* Get the input image data. */
	startclk = clock();
	if (!(image = image_load(in, cmdopts->infmt, cmdopts->inopts))) {
		fprintf(stderr, "cannot load image data\n");
		exit(EXIT_FAILURE);
	}
	endclk = clock();
	dectime = endclk - startclk;

	/* If requested, throw away all of the components except one.
	  Why might this be desirable?  It is a hack, really.
	  None of the image formats other than the JPEG-2000 ones support
	  images with two, four, five, or more components.  This hack
	  allows such images to be decoded with the non-JPEG-2000 decoders,
	  one component at a time. */
	numcmpts = image_numcmpts(image);
	if (cmdopts->cmptno >= 0 && cmdopts->cmptno < numcmpts) {
		for (i = numcmpts - 1; i >= 0; --i) {
			if (i != cmdopts->cmptno) {
				image_delcmpt(image, i);
			}
		}
	}

	/* Generate the output image data. */
	startclk = clock();
	if (image_save(image, out, cmdopts->outfmt, cmdopts->outopts)) {
		fprintf(stderr, "cannot encode image\n");
		exit(EXIT_FAILURE);
	}
	stream_flush(out);
	endclk = clock();
	enctime = endclk - startclk;

	if (cmdopts->verbose) {
		fprintf(stderr, "decoding time = %f\n", dectime / (double)
		  CLOCKS_PER_SEC);
		fprintf(stderr, "encoding time = %f\n", enctime / (double)
		  CLOCKS_PER_SEC);
	}

	/* If this fails, we don't care. */
	(void) stream_close(in);

	/* Close the output image stream. */
	if (stream_close(out)) {
		fprintf(stderr, "error closing output image file\n");
		exit(EXIT_FAILURE);
	}

	cmdopts_destroy(cmdopts);
	image_destroy(image);
	image_clearfmts();

	/* Success at last! :-) */
	return EXIT_SUCCESS;
}

cmdopts_t *cmdopts_parse(int argc, char **argv)
{

	typedef enum {
		CMDOPT_HELP = 0,
		CMDOPT_TERSE,
		CMDOPT_INFILE,
		CMDOPT_INFMT,
		CMDOPT_INOPT,
		CMDOPT_OUTFILE,
		CMDOPT_OUTFMT,
		CMDOPT_OUTOPT,
		CMDOPT_VERSION,
		CMDOPT_CMPTNO
	} cmdoptid_t;

	static opt_t cmdoptions[] = {
		{CMDOPT_HELP, "help", 0},
		{CMDOPT_TERSE, "terse", 0},
		{CMDOPT_TERSE, "q", 0},
		{CMDOPT_INFILE, "input", OPT_HASARG},
		{CMDOPT_INFILE, "f", OPT_HASARG},
		{CMDOPT_INFMT, "input-format", OPT_HASARG},
		{CMDOPT_INFMT, "t", OPT_HASARG},
		{CMDOPT_INOPT, "input-option", OPT_HASARG},
		{CMDOPT_INOPT, "o", OPT_HASARG},
		{CMDOPT_OUTFILE, "output", OPT_HASARG},
		{CMDOPT_OUTFILE, "F", OPT_HASARG},
		{CMDOPT_OUTFMT, "output-format", OPT_HASARG},
		{CMDOPT_OUTFMT, "T", OPT_HASARG},
		{CMDOPT_OUTOPT, "output-option", OPT_HASARG},
		{CMDOPT_OUTOPT, "O", OPT_HASARG},
		{CMDOPT_VERSION, "version", 0},
		{CMDOPT_CMPTNO, "cmptno", OPT_HASARG},
		{-1, 0, 0}
	};

	cmdopts_t *cmdopts;
	int c;

	if (!(cmdopts = malloc(sizeof(cmdopts_t)))) {
		fprintf(stderr, "no memory\n");
		exit(EXIT_FAILURE);
	}

	cmdopts->infile = 0;
	cmdopts->infmt = -1;
	cmdopts->inopts = 0;
	cmdopts->inoptsbuf[0] = '\0';
	cmdopts->outfile = 0;
	cmdopts->outfmt = -1;
	cmdopts->outopts = 0;
	cmdopts->outoptsbuf[0] = '\0';
	cmdopts->verbose = 1;
	cmdopts->version = 0;
	cmdopts->cmptno = -1;

	while ((c = getoptx(argc, argv, cmdoptions)) != EOF) {
		switch (c) {
		case CMDOPT_HELP:
		default:
			cmdusage();
			break;
		case CMDOPT_TERSE:
			cmdopts->verbose = 0;
			break;
		case CMDOPT_VERSION:
			cmdopts->version = 1;
			cmdopts->verbose = 0;
			break;
		case CMDOPT_INFILE:
			cmdopts->infile = optarg;
			break;
		case CMDOPT_INFMT:
			if ((cmdopts->infmt = image_strtofmt(optarg)) < 0) {
				fprintf(stderr, "warning: invalid format %s\n", optarg);
				cmdusage();
			}
			break;
		case CMDOPT_INOPT:
			addopt(cmdopts->inoptsbuf, OPTSMAX, optarg);
			cmdopts->inopts = cmdopts->inoptsbuf;
			break;
		case CMDOPT_OUTFILE:
			cmdopts->outfile = optarg;
			break;
		case CMDOPT_OUTFMT:
			if ((cmdopts->outfmt = image_strtofmt(optarg)) < 0) {
				fprintf(stderr, "warning: invalid format %s\n",
				  optarg);
				cmdusage();
			}
			break;
		case CMDOPT_OUTOPT:
			addopt(cmdopts->outoptsbuf, OPTSMAX, optarg);
			cmdopts->outopts = cmdopts->outoptsbuf;
			break;
		case CMDOPT_CMPTNO:
			cmdopts->cmptno = atoi(optarg);
			break;
		}
	}

	if (cmdopts->infmt < 0 && cmdopts->infile) {
		cmdopts->infmt = image_fmtfromname(cmdopts->infile);
	}

	if (cmdopts->outfmt < 0 && cmdopts->outfile) {
		cmdopts->outfmt = image_fmtfromname(cmdopts->outfile);
	}

	return cmdopts;
}

void cmdopts_destroy(cmdopts_t *cmdopts)
{
	free(cmdopts);
}

int addopt(char *optstr, int maxlen, char *s)
{
	int n;
	int m;

	n = strlen(optstr);
	m = n + strlen(s) + 1;
	if (m > maxlen) {
		return 1;
	}
	if (n > 0) {
		strcat(optstr, "\n");
	}
	strcat(optstr, s);
	return 0;
}

void cmdinfo()
{
	fprintf(stderr, "JasPer Transcoder (Version %s).\n",
	  VERSION);
	fprintf(stderr, "%s\n", COPYRIGHT);
	fprintf(stderr, "%s\n", MISCNOTES);
}

static char helpinfo[] =
  "The following options are supported:\n"
  "    --help\n"
  "        Print this usage information and exit.\n"
  "    --version\n"
  "        Print version information to standard output and exit.\n"
  "    --input FILE\n"
  "    -f FILE\n"
  "        Read the input image from the file named FILE.\n"
  "        By default, the input image is read from standard input.\n"
  "    --input-format FORMAT\n"
  "    -t FORMAT\n"
  "        Specify the format of the input image as FORMAT\n"
  "        (e.g., bmp, jpc, pnm, ras).\n"
  "    --input-option OPTION\n"
  "    -o OPTION\n"
  "        Provide the option OPTION to the decoder.\n"
  "    --output FILE\n"
  "    -F FILE\n"
  "        Write the encoded image to the file named FILE.\n"
  "        By default, the encoded image is written to standard output.\n"
  "    --output-format FORMAT\n"
  "    -T FORMAT\n"
  "        Specify the format of the output image as FORMAT\n"
  "        (e.g., bmp, jpc, pnm, ras).\n"
  "    --output-option OPTION\n"
  "    -O OPTION\n"
  "        Provide the option OPTION to the encoder.\n"
  ;

void cmdusage()
{
	cmdinfo();
	fprintf(stderr, "usage: %s [options]\n", cmdname);
	fprintf(stderr, "%s", helpinfo);
	exit(EXIT_FAILURE);
}
