/*
 * Copyright (c) 1999-2000 Image Power, Inc. and the University of
 *   British Columbia.
 * Copyright (c) 2001 Michael David Adams.
 * All rights reserved.
 *
 * The complete licensing terms for this software can be found in the file
 * named "LICENSE" in this software distribution.
 */

/*
 * Image Library
 *
 * $Id$
 */

/******************************************************************************\
* Includes.
\******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>

#include "mathx.h"
#include "image.h"
#include "memalloc.h"
#include "stringx.h"

/******************************************************************************\
* Types.
\******************************************************************************/

/******************************************************************************\
* Local prototypes.
\******************************************************************************/

static void image_cmpt_destroy(image_cmpt_t *cmpt);
static image_cmpt_t *image_cmpt_create(uint_fast32_t tlx, uint_fast32_t tly,
  uint_fast32_t hstep, uint_fast32_t vstep, uint_fast32_t width, uint_fast32_t
  height, uint_fast16_t depth, bool sgnd, uint_fast32_t inmem);
static image_fmtinfo_t *image_lookupfmtbyid(int id);
static image_fmtinfo_t *image_lookupfmtbyname(const char *name);
static void image_setbbox(image_t *image);
static image_cmpt_t *image_cmpt_copy(image_cmpt_t *cmpt);
static image_cmpt_t *image_cmpt_create0();
static int image_growcmpts(image_t *image, int maxcmpts);
static uint_fast32_t inttobits(fix_t v, int prec, bool sgnd);
static fix_t bitstoint(uint_fast32_t v, int prec, bool sgnd);

/******************************************************************************\
* Global data.
\******************************************************************************/

static int image_numfmts = 0;
static image_fmtinfo_t image_fmtinfos[IMAGE_MAXFMTS];

/******************************************************************************\
* Create and destroy operations.
\******************************************************************************/

image_t *image_create(uint_fast16_t numcmpts, image_cmptparm_t *cmptparms,
  int colormodel)
{
	image_t *image;
	uint_fast32_t rawsize;
	uint_fast32_t inmem;
	uint_fast16_t cmptno;
	image_cmptparm_t *cmptparm;

	if (!(image = malloc(sizeof(image_t)))) {
		return 0;
	}

	image->tlx_ = 0;
	image->tly_ = 0;
	image->brx_ = 0;
	image->bry_ = 0;
	image->colormodel_ = colormodel;
	image->numcmpts_ = 0;
	image->maxcmpts_ = numcmpts;
	image->cmpts_ = 0;
image->inmem_ = true;

	/* Allocate memory for the per-component information. */
	if (!(image->cmpts_ = malloc(image->maxcmpts_ *
	  sizeof(image_cmpt_t *)))) {
		image_destroy(image);
		return 0;
	}
	/* Initialize in case of failure. */
	for (cmptno = 0; cmptno < image->maxcmpts_; ++cmptno) {
		image->cmpts_[cmptno] = 0;
	}

	/* Compute the approximate raw size of the image. */
	rawsize = 0;
	for (cmptno = 0, cmptparm = cmptparms; cmptno < numcmpts; ++cmptno,
	  ++cmptparm) {
		rawsize += (cmptparm->width * cmptparm->height *
		  cmptparm->prec + 7) / 8;
	}
	/* Decide whether to buffer the image data in memory, based on the
	  raw size of the image. */
	inmem = (rawsize < IMAGE_INMEMTHRESH);

	/* Create the individual image components. */
	for (cmptno = 0, cmptparm = cmptparms; cmptno < numcmpts; ++cmptno,
	  ++cmptparm) {
		if (!(image->cmpts_[cmptno] = image_cmpt_create(cmptparm->tlx,
		  cmptparm->tly, cmptparm->hstep, cmptparm->vstep,
		  cmptparm->width, cmptparm->height, cmptparm->prec,
		  cmptparm->sgnd, inmem))) {
			image_destroy(image);
			return 0;
		}
		++image->numcmpts_;
	}

	/* Determine the bounding box for all of the components on the
	  reference grid (i.e., the image area) */
	image_setbbox(image);

	return image;
}

image_t *image_create0()
{
	image_t *image;

	if (!(image = malloc(sizeof(image_t)))) {
		return 0;
	}

	image->tlx_ = 0;
	image->tly_ = 0;
	image->brx_ = 0;
	image->bry_ = 0;
	image->colormodel_ = CM_UNKNOWN;
	image->numcmpts_ = 0;
	image->maxcmpts_ = 0;
	image->cmpts_ = 0;
	image->inmem_ = true;

	return image;
}

image_t *image_copy(image_t *image)
{
	image_t *newimage;
	int cmptno;

	newimage = image_create0();
	if (image_growcmpts(newimage, image->numcmpts_)) {
		goto error;
	}
	for (cmptno = 0; cmptno < image->numcmpts_; ++cmptno) {
		if (!(newimage->cmpts_[cmptno] = image_cmpt_copy(image->cmpts_[cmptno]))) {
			goto error;
		}
		++newimage->numcmpts_;
	}

	image_setbbox(newimage);

	return newimage;
error:
	if (newimage) {
		image_destroy(newimage);
	}
	return 0;
}

static image_cmpt_t *image_cmpt_create0()
{
	image_cmpt_t *cmpt;
	if (!(cmpt = malloc(sizeof(image_cmpt_t)))) {
		return 0;
	}
	memset(cmpt, 0, sizeof(image_cmpt_t));
	return cmpt;
}

static image_cmpt_t *image_cmpt_copy(image_cmpt_t *cmpt)
{
	image_cmpt_t *newcmpt;

	if (!(newcmpt = image_cmpt_create0())) {
		return 0;
	}
	newcmpt->tlx_ = cmpt->tlx_;
	newcmpt->tly_ = cmpt->tly_;
	newcmpt->hstep_ = cmpt->hstep_;
	newcmpt->vstep_ = cmpt->vstep_;
	newcmpt->width_ = cmpt->width_;
	newcmpt->height_ = cmpt->height_;
	newcmpt->prec_ = cmpt->prec_;
	newcmpt->sgnd_ = cmpt->sgnd_;
	newcmpt->cps_ = cmpt->cps_;
	if (!(newcmpt->stream_ = stream_memopen(0, 0))) {
		return 0;
	}
	if (stream_seek(cmpt->stream_, 0, SEEK_SET)) {
		return 0;
	}
	if (stream_copy(newcmpt->stream_, cmpt->stream_, -1)) {
		return 0;
	}
	if (stream_seek(newcmpt->stream_, 0, SEEK_SET)) {
		return 0;
	}
	return newcmpt;
}

void image_destroy(image_t *image)
{
	int i;

	if (image->cmpts_) {
		for (i = 0; i < image->numcmpts_; ++i) {
			image_cmpt_destroy(image->cmpts_[i]);
			image->cmpts_[i] = 0;
		}
		free(image->cmpts_);
	}
	free(image);
}

static image_cmpt_t *image_cmpt_create(uint_fast32_t tlx, uint_fast32_t tly,
  uint_fast32_t hstep, uint_fast32_t vstep, uint_fast32_t width, uint_fast32_t
  height, uint_fast16_t depth, bool sgnd, uint_fast32_t inmem)
{
	image_cmpt_t *cmpt;
	long size;

	if (!(cmpt = malloc(sizeof(image_cmpt_t)))) {
		return 0;
	}

	cmpt->tlx_ = tlx;
	cmpt->tly_ = tly;
	cmpt->hstep_ = hstep;
	cmpt->vstep_ = vstep;
	cmpt->width_ = width;
	cmpt->height_ = height;
	cmpt->prec_ = depth;
	cmpt->sgnd_ = sgnd;
	cmpt->stream_ = 0;
	cmpt->cps_ = (depth + 7) / 8;

	size = cmpt->width_ * cmpt->height_ * cmpt->cps_;
	cmpt->stream_ = (inmem) ? stream_memopen(0, size) : stream_tmpfile();
	if (!cmpt->stream_) {
		image_cmpt_destroy(cmpt);
		return 0;
	}

	/* Zero the component data.  This isn't necessary, but it is
	convenient for debugging purposes. */
	if (stream_seek(cmpt->stream_, size - 1, SEEK_SET) < 0 ||
	  stream_putc(cmpt->stream_, 0) == EOF ||
	  stream_seek(cmpt->stream_, 0, SEEK_SET) < 0) {
		image_cmpt_destroy(cmpt);
		return 0;
	}

	return cmpt;
}

static void image_cmpt_destroy(image_cmpt_t *cmpt)
{
	if (cmpt->stream_) {
		stream_close(cmpt->stream_);
	}
	free(cmpt);
}

/******************************************************************************\
* Load and save operations.
\******************************************************************************/

image_t *image_load(stream_t *in, int fmt, char *optstr)
{
	image_fmtinfo_t *fmtinfo;

	/* If possible, try to determine the format of the input data. */
	if (fmt < 0) {
		if ((fmt = image_getfmt(in)) < 0) {
			return 0;
		}
	}
	if (!(fmtinfo = image_lookupfmtbyid(fmt))) {
		return 0;
	}
	return (fmtinfo->ops.load) ? (*fmtinfo->ops.load)(in, optstr) : 0;
}

int image_save(image_t *image, stream_t *out, int fmt, char *optstr)
{
	image_fmtinfo_t *fmtinfo;
	if (!(fmtinfo = image_lookupfmtbyid(fmt))) {
		return -1;
	}
	return (fmtinfo->ops.save) ? (*fmtinfo->ops.save)(image, out,
	  optstr) : (-1);
}

/******************************************************************************\
* Component read and write operations.
\******************************************************************************/

int image_readcmpt(image_t *image, uint_fast16_t cmptno, uint_fast32_t x, uint_fast32_t y, uint_fast32_t width,
  uint_fast32_t height, matrix_t *data)
{
	image_cmpt_t *cmpt;
	uint_fast32_t i;
	uint_fast32_t j;
	int k;
	fix_t v;
	int c;
	fix_t *dr;
	fix_t *d;
	int drs;

	if (cmptno < 0 || cmptno >= image->numcmpts_) {
		return -1;
	}

	cmpt = image->cmpts_[cmptno];
	if (x >= cmpt->width_ || y >= cmpt->height_ ||
	  x + width > cmpt->width_ ||
	  y + height > cmpt->height_) {
		return -1;
	}

	if (matrix_numrows(data) != height || matrix_numcols(data) != width) {
		if (matrix_resize(data, height, width)) {
			return -1;
		}
	}

	dr = matrix_getref(data, 0, 0);
	drs = matrix_rowstep(data);
	for (i = 0; i < height; ++i, dr += drs) {
		d = dr;
		if (stream_seek(cmpt->stream_, (cmpt->width_ * (y + i) + x)
		  * cmpt->cps_, SEEK_SET) < 0) {
			return -1;
		}
		for (j = width; j > 0; --j, ++d) {
			v = 0;
			for (k = cmpt->cps_; k > 0; --k) {
				if ((c = stream_getc(cmpt->stream_)) == EOF) {
					return -1;
				}
				v = (v << 8) | (c & 0xff);
			}
			*d = bitstoint(v, cmpt->prec_, cmpt->sgnd_);
		}
	}

	return 0;
}

int image_writecmpt(image_t *image, uint_fast16_t cmptno, uint_fast32_t x, uint_fast32_t y, uint_fast32_t width,
  uint_fast32_t height, matrix_t *data)
{
	image_cmpt_t *cmpt;
	uint_fast32_t i;
	uint_fast32_t j;
	fix_t *d;
	fix_t *dr;
	int drs;
	fix_t v;
	int k;
	int c;

	if (cmptno < 0 || cmptno >= image->numcmpts_) {
		return -1;
	}

	cmpt = image->cmpts_[cmptno];
	if (x >= cmpt->width_ || y >= cmpt->height_ ||
	  x + width > cmpt->width_ ||
	  y + height > cmpt->height_) {
		return -1;
	}

	if (matrix_numrows(data) != height || matrix_numcols(data) != width) {
		return -1;
	}

	dr = matrix_getref(data, 0, 0);
	drs = matrix_rowstep(data);
	for (i = 0; i < height; ++i, dr += drs) {
		d = dr;
		if (stream_seek(cmpt->stream_, (cmpt->width_ * (y + i) + x)
		  * cmpt->cps_, SEEK_SET) < 0) {
			return -1;
		}
		for (j = width; j > 0; --j, ++d) {
			v = inttobits(*d, cmpt->prec_, cmpt->sgnd_);
			for (k = cmpt->cps_; k > 0; --k) {
				c = (v >> (8 * (cmpt->cps_ - 1))) & 0xff;
				if (stream_putc(cmpt->stream_,
				  (unsigned char) c) == EOF) {
					return -1;
				}
				v <<= 8;
			}
		}
	}

	return 0;
}

/******************************************************************************\
* File format operations.
\******************************************************************************/

void image_clearfmts()
{
	int i;
	image_fmtinfo_t *fmtinfo;
	for (i = 0; i < image_numfmts; ++i) {
		fmtinfo = &image_fmtinfos[i];
		if (fmtinfo->name) {
			free(fmtinfo->name);
			fmtinfo->name = 0;
		}
		if (fmtinfo->ext) {
			free(fmtinfo->ext);
			fmtinfo->ext = 0;
		}
	}
}

int image_addfmt(int id, char *name, char *ext, image_fmtops_t *ops)
{
	image_fmtinfo_t *fmtinfo;
	assert(id >= 0 && name && ext && ops);
	if (image_numfmts >= IMAGE_MAXFMTS) {
		return -1;
	}
	fmtinfo = &image_fmtinfos[image_numfmts];
	fmtinfo->id = id;
	if (!(fmtinfo->name = strclone(name))) {
		return -1;
	}
	if (!(fmtinfo->ext = strclone(ext))) {
		free(fmtinfo->name);
		return -1;
	}
	fmtinfo->ops = *ops;
	++image_numfmts;
	return 0;
}

int image_strtofmt(char *name)
{
	image_fmtinfo_t *fmtinfo;
	if (!(fmtinfo = image_lookupfmtbyname(name))) {
		return -1;
	}
	return fmtinfo->id;
}

char *image_fmttostr(int fmt)
{
	image_fmtinfo_t *fmtinfo;
	if (!(fmtinfo = image_lookupfmtbyid(fmt))) {
		return 0;
	}
	return fmtinfo->name;
}

int image_getfmt(stream_t *in)
{
	image_fmtinfo_t *fmtinfo;
	int found;
	int i;

	/* Check for data in each of the supported formats. */
	found = 0;
	for (i = 0, fmtinfo = image_fmtinfos; i < image_numfmts; ++i,
	  ++fmtinfo) {
		if (fmtinfo->ops.validate) {
			/* Is the input data valid for this format? */
			if (!(*fmtinfo->ops.validate)(in)) {
				found = 1;
				break;
			}
		}
	}
	return found ? fmtinfo->id : (-1);
}

int image_fmtfromname(char *name)
{
	int i;
	char *ext;
	image_fmtinfo_t *fmtinfo;
	/* Get the file name extension. */
	if (!(ext = strrchr(name, '.'))) {
		return -1;
	}
	++ext;
	/* Try to find a format that uses this extension. */	
	for (i = 0, fmtinfo = image_fmtinfos; i < image_numfmts; ++i,
	  ++fmtinfo) {
		/* Do we have a match? */
		if (!strcmp(ext, fmtinfo->ext)) {
			return fmtinfo->id;
		}
	}
	return -1;
}

/******************************************************************************\
* Miscellaneous operations.
\******************************************************************************/

uint_fast32_t image_getrawsize(image_t *image)
{
	uint_fast32_t rawsize;
	uint_fast32_t cmptno;
	image_cmpt_t *cmpt;

	rawsize = 0;
	for (cmptno = 0; cmptno < image->numcmpts_; ++cmptno) {
		cmpt = image->cmpts_[cmptno];
		rawsize += (cmpt->width_ * cmpt->height_ * cmpt->prec_ +
		  7) / 8;
	}
	return rawsize;
}

void image_delcmpt(image_t *image, uint_fast16_t cmptno)
{
	if (cmptno >= image->numcmpts_) {
		return;
	}
	image_cmpt_destroy(image->cmpts_[cmptno]);
	if (cmptno < image->numcmpts_) {
		memmove(&image->cmpts_[cmptno], &image->cmpts_[cmptno + 1],
		  (image->numcmpts_ - 1 - cmptno) * sizeof(image_cmpt_t *));
	}
	--image->numcmpts_;

	image_setbbox(image);
}

int image_addcmpt(image_t *image, uint_fast16_t cmptno,
  image_cmptparm_t *cmptparm)
{
	image_cmpt_t *newcmpt;
	assert(cmptno >= 0 && cmptno <= image->numcmpts_);
	if (image->numcmpts_ >= image->maxcmpts_) {
		if (image_growcmpts(image, image->maxcmpts_ + 128)) {
			return -1;
		}
	}
	if (!(newcmpt = image_cmpt_create(cmptparm->tlx,
	  cmptparm->tly, cmptparm->hstep, cmptparm->vstep,
	  cmptparm->width, cmptparm->height, cmptparm->prec,
	  cmptparm->sgnd, 1))) {
		return -1;
	}
	if (cmptno < image->numcmpts_) {
		memmove(&image->cmpts_[cmptno + 1], &image->cmpts_[cmptno],
		  (image->numcmpts_ - cmptno) * sizeof(image_cmpt_t *));
	}
	image->cmpts_[cmptno] = newcmpt;
	++image->numcmpts_;

	image_setbbox(image);

	return 0;
}

static image_fmtinfo_t *image_lookupfmtbyid(int id)
{
	int i;
	image_fmtinfo_t *fmtinfo;

	for (i = 0, fmtinfo = image_fmtinfos; i < image_numfmts; ++i, ++fmtinfo) {
		if (fmtinfo->id == id) {
			return fmtinfo;
		}
	}
	return 0;
}

static image_fmtinfo_t *image_lookupfmtbyname(const char *name)
{
	int i;
	image_fmtinfo_t *fmtinfo;

	for (i = 0, fmtinfo = image_fmtinfos; i < image_numfmts; ++i, ++fmtinfo) {
		if (!strcmp(fmtinfo->name, name)) {
			return fmtinfo;
		}
	}
	return 0;
}





static uint_fast32_t inttobits(fix_t v, int prec, bool sgnd)
{
	uint_fast32_t ret;
	ret = ((sgnd && v < 0) ? ((1 << prec) + v) : v) & ONES(prec);
if (sgnd && 0) {
fprintf(stderr, "(%d,%d,%d)->(0x%x)...", v, prec,sgnd,ret);
}
	return ret;
}

static fix_t bitstoint(uint_fast32_t v, int prec, bool sgnd)
{
	fix_t ret;
	v &= ONES(prec);
	ret = (sgnd && (v & (1 << (prec - 1)))) ? (v - (1 << prec)) : v;
if (sgnd && 0) {
fprintf(stderr, "(0x%x,%d,%d)->(%d)...", v, prec,sgnd,ret);
}
	return ret;
}

static void image_setbbox(image_t *image)
{
	image_cmpt_t *cmpt;
	int cmptno;
	int_fast32_t x;
	int_fast32_t y;

	if (image->numcmpts_ > 0) {
		/* Determine the bounding box for all of the components on the
		  reference grid (i.e., the image area) */
		cmpt = image->cmpts_[0];
		image->tlx_ = cmpt->tlx_;
		image->tly_ = cmpt->tly_;
		image->brx_ = cmpt->tlx_ + cmpt->hstep_ * (cmpt->width_ - 1) + 1;
		image->bry_ = cmpt->tly_ + cmpt->vstep_ * (cmpt->height_ - 1) + 1;
		for (cmptno = 1; cmptno < image->numcmpts_; ++cmptno) {
			cmpt = image->cmpts_[cmptno];
			if (image->tlx_ > cmpt->tlx_) {
				image->tlx_ = cmpt->tlx_;
			}
			if (image->tly_ > cmpt->tly_) {
				image->tly_ = cmpt->tly_;
			}
			x = cmpt->tlx_ + cmpt->hstep_ * (cmpt->width_ - 1) + 1;
			if (image->brx_ < x) {
				image->brx_ = x;
			}
			y = cmpt->tly_ + cmpt->vstep_ * (cmpt->height_ - 1) + 1;
			if (image->bry_ < y) {
				image->bry_ = y;
			}
		}
	} else {
		image->tlx_ = 0;
		image->tly_ = 0;
		image->brx_ = 0;
		image->bry_ = 0;
	}
}

static int image_growcmpts(image_t *image, int maxcmpts)
{
	image_cmpt_t **newcmpts;
	int cmptno;

	newcmpts = (!image->cmpts_) ? malloc(maxcmpts * sizeof(image_cmpt_t *)) :
	  realloc(image->cmpts_, maxcmpts * sizeof(image_cmpt_t *));
	if (!newcmpts) {
		return -1;
	}
	image->cmpts_ = newcmpts;
	image->maxcmpts_ = maxcmpts;
	for (cmptno = image->numcmpts_; cmptno < image->maxcmpts_; ++cmptno) {
		image->cmpts_[cmptno] = 0;
	}
	return 0;
}

int image_copycmpt(image_t *dstimage, int dstcmptno, image_t *srcimage,
  int srccmptno)
{
	image_cmpt_t *newcmpt;
	if (dstimage->numcmpts_ >= dstimage->maxcmpts_) {
		if (image_growcmpts(dstimage, dstimage->maxcmpts_ + 128)) {
			return -1;
		}
	}
	if (!(newcmpt = image_cmpt_copy(srcimage->cmpts_[srccmptno]))) {
		return -1;
	}
	if (dstcmptno < dstimage->numcmpts_) {
		memmove(&dstimage->cmpts_[dstcmptno + 1], &dstimage->cmpts_[dstcmptno],
		  (dstimage->numcmpts_ - dstcmptno) * sizeof(image_cmpt_t *));
	}
	dstimage->cmpts_[dstcmptno] = newcmpt;
	++dstimage->numcmpts_;

	image_setbbox(dstimage);
	return 0;
}


void image_dump(image_t *image, FILE *out)
{
	int cmptno;
	seq2d_t *data;
	image_cmpt_t *cmpt;
	if (!(data = seq2d_create(0, 0, 1, 1))) {
		abort();
	}
	for (cmptno = 0; cmptno < image->numcmpts_; ++cmptno) {
		cmpt = image->cmpts_[cmptno];
		fprintf(out, "prec=%d sgnd=%d\n", cmpt->prec_, cmpt->sgnd_);
		if (image_readcmpt(image, cmptno, 0, 0, 1, 1, data)) {
			abort();
		}
		fprintf(out, "tlsample %ld\n", (long) seq2d_get(data, 0, 0));
	}
	seq2d_destroy(data);
}
