/*
 * Copyright (c) 2001 Michael David Adams.
 * All rights reserved.
 *
 * The complete licensing terms for this software can be found in the file
 * named "LICENSE" in this software distribution.
 */

/*
 * Tag-Value Parser Library
 *
 * $Id$
 */

/******************************************************************************\
* Includes.
\******************************************************************************/

#include <assert.h>
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>

#include "memalloc.h"
#include "stringx.h"
#include "tvparser.h"

/******************************************************************************\
* Macros.
\******************************************************************************/

/* Is the specified character valid for a tag name? */
#define	isidentifier(x) \
	(isalpha(x) || (x) == '_' || isdigit(x))

/******************************************************************************\
* Code for creating and destroying a tag-value parser.
\******************************************************************************/

tvparser_t *tvparser_create(const char *s)
{
	tvparser_t *tvp;
	if (!(tvp = malloc(sizeof(tvparser_t)))) {
		return 0;
	}
	if (!(tvp->buf = strclone(s))) {
		tvparser_destroy(tvp);
		return 0;
	}
	tvp->pos = tvp->buf;
	tvp->tag = 0;
	tvp->val = 0;
	return tvp;
}

void tvparser_destroy(tvparser_t *tvp)
{
	if (tvp->buf) {
		free(tvp->buf);
	}
	free(tvp);
}

/******************************************************************************\
* Main parsing code.
\******************************************************************************/

/* Get the next tag-value pair. */
int tvparser_next(tvparser_t *tvp)
{
	char *p;
	char *tag;
	char *val;

	/* Skip any leading whitespace. */
	p = tvp->pos;
	while (*p != '\0' && isspace(*p)) {
		++p;
	}

	/* Has the end of the input data been reached? */
	if (*p == '\0') {
		/* No more tags are present. */
		tvp->pos = p;
		return 1;
	}

	/* Does the tag name begin with a valid character? */
	if (!isidentifier(*p)) {
		return -1;
	}

	/* Remember where the tag name begins. */
	tag = p;

	/* Find the end of the tag name. */
	while (*p != '\0' && isidentifier(*p)) {
		++p;
	}

	/* Has the end of the input data been reached? */
	if (*p == '\0') {
		/* The value field is empty. */
		tvp->tag = tag;
		tvp->val = "";
		tvp->pos = p;
		return 0;
	}

	/* Is a value field not present? */
	if (*p != '=') {
		if (*p != '\0' && !isspace(*p)) {
			return -1;
		}
		*p++ = '\0';
		tvp->tag = tag;
		tvp->val = "";
		tvp->pos = p;
		return 0;
	}

	*p++ = '\0';

	val = p;
	while (*p != '\0' && !isspace(*p)) {
		++p;
	}

	if (*p != '\0') {
		*p++ = '\0';
	}

	tvp->pos = p;
	tvp->tag = tag;
	tvp->val = val;

	return 0;
}

/******************************************************************************\
* Code for querying the current tag/value.
\******************************************************************************/

/* Get the current tag. */
char *tvparser_gettag(tvparser_t *tvp)
{
	return tvp->tag;
}

/* Get the current value. */
char *tvparser_getval(tvparser_t *tvp)
{
	return tvp->val;
}

/******************************************************************************\
* Miscellaneous code.
\******************************************************************************/

/* Lookup a tag by name. */
taginfo_t *taginfos_lookup(taginfo_t *taginfos, const char *name)
{
	taginfo_t *taginfo;
	taginfo = taginfos;
	while (taginfo->id >= 0) {
		if (!strcmp(taginfo->name, name)) {
			return taginfo;
		}
		++taginfo;
	}
	return 0;
}

/* This function is simply for convenience. */
/* One can avoid testing for the special case of a null pointer, by
  using this function.   This function never returns a null pointer.  */
taginfo_t *taginfo_nonull(taginfo_t *taginfo)
{
	static taginfo_t invalidtaginfo = {
		-1, 0
	};
	
	return taginfo ? taginfo : &invalidtaginfo;
}
