/*
 * Copyright (c) 1999-2000 Image Power, Inc. and the University of
 *   British Columbia.
 * Copyright (c) 2001 Michael David Adams.
 * All rights reserved.
 *
 * The complete licensing terms for this software can be found in the file
 * named "LICENSE" in this software distribution.
 */

/*
 * MQ Arithmetic Decoder
 *
 * $Id$
 */

/******************************************************************************\
* Includes.
\******************************************************************************/

#include "types.h"
#include <assert.h>
#include <stdlib.h>
#include <stdarg.h>

#include "mqdec.h"
#include "memalloc.h"
#include "util.h"

/******************************************************************************\
* Local macros.
\******************************************************************************/

#if defined(DEBUG)
#define	MQDEC_CALL(n, x) \
	((getdbglevel() >= (n)) ? ((void)(x)) : ((void)0))
#else
#define	MQDEC_CALL(n, x)
#endif

/******************************************************************************\
* Local function prototypes.
\******************************************************************************/

static void mqdec_bytein(mqdec_t *mqdec);

/******************************************************************************\
* Code for creation and destruction of a MQ decoder.
\******************************************************************************/

/* Create a MQ decoder. */
mqdec_t *mqdec_create(int maxctxs, stream_t *in)
{
	mqdec_t *mqdec;

	/* There must be at least one context. */
	assert(maxctxs > 0);

	/* Allocate memory for the MQ decoder. */
	if (!(mqdec = malloc(sizeof(mqdec_t)))) {
		goto error;
	}
	mqdec->in = in;
	mqdec->maxctxs = maxctxs;
	/* Allocate memory for the per-context state information. */
	if (!(mqdec->ctxs = malloc(mqdec->maxctxs * sizeof(mqstate_t *)))) {
		goto error;
	}
	/* Set the current context to the first context. */
	mqdec->curctx = mqdec->ctxs;

	/* If an input stream has been associated with the MQ decoder,
	  initialize the decoder state from the stream. */
	if (mqdec->in) {
		mqdec_init(mqdec);
	}
	/* Initialize the per-context state information. */
	mqdec_setctxs(mqdec, 0, 0);

	return mqdec;

error:
	/* Oops...  Something has gone wrong. */
	if (mqdec) {
		mqdec_destroy(mqdec);
	}
	return 0;
}

/* Destroy a MQ decoder. */
void mqdec_destroy(mqdec_t *mqdec)
{
	if (mqdec->ctxs) {
		free(mqdec->ctxs);
	}
	free(mqdec);
}

/******************************************************************************\
* Code for initialization of a MQ decoder.
\******************************************************************************/

void mqdec_init(mqdec_t *mqdec)
{
	int c;

	mqdec->eof = 0;
	mqdec->creg = 0;
	/* Get the next byte from the input stream. */
	if ((c = stream_getc(mqdec->in)) == EOF) {
		/* We have encountered an I/O error or EOF. */
		c = 0xff;
		mqdec->eof = 1;
	}
	mqdec->inbuffer = c;
	mqdec->creg += mqdec->inbuffer << 16;
	mqdec_bytein(mqdec);
	mqdec->creg <<= 7;
	mqdec->ctreg -= 7;
	mqdec->areg = 0x8000;
}

void mqdec_setinput(mqdec_t *mqdec, stream_t *in)
{
	mqdec->in = in;
}

void mqdec_setctxs(mqdec_t *mqdec, int numctxs, mqctx_t *ctxs)
{
	mqstate_t **ctx;
	int n;

	ctx = mqdec->ctxs;
	n = min(mqdec->maxctxs, numctxs);
	while (--n >= 0) {
		*ctx = &mqstates[2 * ctxs->ind + ctxs->mps];
		++ctx;
		++ctxs;
	}
	n = mqdec->maxctxs - numctxs;
	while (--n >= 0) {
		*ctx = &mqstates[0];
		++ctx;
	}
}

void mqdec_setctx(mqdec_t *mqdec, int ctxno, mqctx_t *ctx)
{
	mqstate_t **ctxi;
	ctxi = &mqdec->ctxs[ctxno];
	*ctxi = &mqstates[2 * ctx->ind + ctx->mps];
}

/******************************************************************************\
* Code for decoding a bit.
\******************************************************************************/

int mqdec_getbit_func(register mqdec_t *mqdec)
{
	int bit;
	LOGMSG(100, ("mqdec_getbit_func(%p)\n", mqdec));
	MQDEC_CALL(100, mqdec_dump(mqdec, stderr));
	bit = mqdec_getbit_macro(mqdec);
	MQDEC_CALL(100, mqdec_dump(mqdec, stderr));
	LOGMSG(100, ("ctx = %d, decoded %d\n", mqdec->curctx -
	  mqdec->ctxs, bit));
	return bit;
}

/* Apply MPS_EXCHANGE algorithm (with RENORMD). */
int mqdec_mpsexchrenormd(register mqdec_t *mqdec)
{
	int ret;
	register mqstate_t *state = *mqdec->curctx;
	mqdec_mpsexchange(mqdec->areg, state->qeval, mqdec->curctx, ret);
	mqdec_renormd(mqdec->areg, mqdec->creg, mqdec->ctreg, mqdec->in,
	  mqdec->eof, mqdec->inbuffer);
	return ret;
}

/* Apply LPS_EXCHANGE algorithm (with RENORMD). */
int mqdec_lpsexchrenormd(register mqdec_t *mqdec)
{
	int ret;
	register mqstate_t *state = *mqdec->curctx;
	mqdec_lpsexchange(mqdec->areg, state->qeval, mqdec->curctx, ret);
	mqdec_renormd(mqdec->areg, mqdec->creg, mqdec->ctreg, mqdec->in,
	  mqdec->eof, mqdec->inbuffer);
	return ret;
}

/******************************************************************************\
* Support code.
\******************************************************************************/

/* Apply the BYTEIN algorithm. */
static void mqdec_bytein(mqdec_t *mqdec)
{
	int c;
	unsigned char prevbuf;

	if (!mqdec->eof) {
		if ((c = stream_getc(mqdec->in)) == EOF) {
			mqdec->eof = 1;
			c = 0xff;
		}
		prevbuf = mqdec->inbuffer;
		mqdec->inbuffer = c;
		if (prevbuf == 0xff) {
			if (c > 0x8f) {
				mqdec->creg += 0xff00;
				mqdec->ctreg = 8;
			} else {
				mqdec->creg += c << 9;
				mqdec->ctreg = 7;
			}
		} else {
			mqdec->creg += c << 8;
			mqdec->ctreg = 8;
		}
	} else {
		mqdec->creg += 0xff00;
		mqdec->ctreg = 8;
	}
}

/******************************************************************************\
* Code for debugging.
\******************************************************************************/

void mqdec_dump(mqdec_t *mqdec, FILE *out)
{
	fprintf(out, "MQDEC A = %08lx, C = %08lx, CT=%08lx, ",
	  (unsigned long) mqdec->areg, (unsigned long) mqdec->creg,
	  (unsigned long) mqdec->ctreg);
	fprintf(out, "CTX = %d, ", mqdec->curctx - mqdec->ctxs);
	fprintf(out, "IND %d, MPS %d, QEVAL %x\n", *mqdec->curctx -
	  mqstates, (*mqdec->curctx)->mps, (*mqdec->curctx)->qeval);
}
