/*
 * Copyright (c) 1999-2000 Image Power, Inc. and the University of
 *   British Columbia.
 * Copyright (c) 2001 Michael David Adams.
 * All rights reserved.
 *
 * The complete licensing terms for this software can be found in the file
 * named "LICENSE" in this software distribution.
 */

/*
 * Portable Pixmap/Graymap Format Support
 *
 * $Id$
 */

/******************************************************************************\
* Includes.
\******************************************************************************/

#include <ctype.h>
#include <math.h>
#include <stdlib.h>
#include <assert.h>

#include "types.h"
#include "pnm.h"
#include "memalloc.h"

/******************************************************************************\
* Local function prototypes.
\******************************************************************************/

static int pnm_gethdr(stream_t *in, pnm_hdr_t *hdr);
static int pnm_getuint16(stream_t *in, uint_fast16_t *val);
static int pnm_getint32(stream_t *in, int_fast32_t *val);
static int pnm_getuint32(stream_t *in, uint_fast32_t *val);
static int pnm_getc(stream_t *in);
static int pnm_getdata(stream_t *in, pnm_hdr_t *hdr, image_t *image);

/******************************************************************************\
* Local data.
\******************************************************************************/

static int pnm_allowtrunc = 1;

/******************************************************************************\
* Load function.
\******************************************************************************/

image_t *image_load_pnm(stream_t *in, char *opts)
{
	pnm_hdr_t hdr;
	image_t *image;
	image_cmptparm_t cmptparms[3];
	image_cmptparm_t *cmptparm;
	int i;
	uint_fast16_t numcmpts;
	int colormodel;

	if (opts) {
		fprintf(stderr, "warning: ignoring options\n");
	}

	/* Read the file header. */
	if (pnm_gethdr(in, &hdr)) {
		return 0;
	}

	/* Determine the number of components and the color model. */
	switch (pnm_type(hdr.magic)) {
	case PNM_TYPE_PGM:
		numcmpts = 1;
		colormodel = CM_GRAY;
		break;
	case PNM_TYPE_PPM:
		numcmpts = 3;
		colormodel = CM_RGB;
		break;
	default:
		/* Prevent annoying compiler warnings about possibly
		  uninitialized variables. */
		numcmpts = 0;
		colormodel = 0;
		abort();
		break;
	}

	/* Create an image of the correct size. */
	for (i = 0, cmptparm = cmptparms; i < numcmpts; ++i, ++cmptparm) {
		cmptparm->tlx = 0;
		cmptparm->tly = 0;
		cmptparm->hstep = 1;
		cmptparm->vstep = 1;
		cmptparm->width = hdr.width;
		cmptparm->height = hdr.height;
		cmptparm->prec = pnm_maxvaltodepth(hdr.maxval);
		cmptparm->sgnd = false;
	}
	if (!(image = image_create(numcmpts, cmptparms, colormodel))) {
		return 0;
	}

	/* Read image data from stream into image. */
	if (pnm_getdata(in, &hdr, image)) {
		image_destroy(image);
		return 0;
	}

	return image;
}

/******************************************************************************\
* Validation function.
\******************************************************************************/

int pnm_validate(stream_t *in)
{
	unsigned char buf[2];
	/* Read the first two characters that constitute the signature. */
	if (stream_read(in, (char *) buf, 2) != 2) {
		return -1;
	}
	/* Put these characters back to the stream. */
	if (stream_ungetc(in, buf[1]) == EOF ||
	  stream_ungetc(in, buf[0]) == EOF) {
		return -1;
	}
	/* Is this the correct signature for a PNM file? */
	if (buf[0] == 'P' && isdigit(buf[1])) {
		return 0;
	}
	return -1;
}

/******************************************************************************\
* Functions for reading the header.
\******************************************************************************/

static int pnm_gethdr(stream_t *in, pnm_hdr_t *hdr)
{
	if (pnm_getuint16(in, &hdr->magic) || pnm_getuint32(in, &hdr->width) ||
	  pnm_getuint32(in, &hdr->height) || pnm_getuint32(in, &hdr->maxval)) {
		return -1;
	}
	return 0;
}

/******************************************************************************\
* Functions for processing the sample data.
\******************************************************************************/

static int pnm_getdata(stream_t *in, pnm_hdr_t *hdr, image_t *image)
{
	int ret;
	int numcmpts;
	uint_fast16_t cmptno;
	int fmt;
	matrix_t *data[3];
	uint_fast32_t x;
	uint_fast32_t y;
	int_fast32_t v;
	int c;

	ret = -1;

	numcmpts = image_numcmpts(image);
	fmt = pnm_fmt(hdr->magic);

	data[0] = 0;
	data[1] = 0;
	data[2] = 0;
	for (cmptno = 0; cmptno < numcmpts; ++cmptno) {
		if (!(data[cmptno] = matrix_create(1, hdr->width))) {
			goto done;
		}
	}

	for (y = 0; y < hdr->height; ++y) {
		for (x = 0; x < hdr->width; ++x) {
			for (cmptno = 0; cmptno < numcmpts; ++cmptno) {
				if (fmt == PNM_FMT_BIN && hdr->maxval <= 255) {
					if ((c = stream_getc(in)) == EOF) {
						if (pnm_allowtrunc) {
							v = 0;
						} else {
							goto done;
						}
					} else {
						v = c;
					}
				} else {
					if (pnm_getint32(in, &v)) {
						if (pnm_allowtrunc) {
							v = 0;
						} else {
							goto done;
						}
					}
				}
				matrix_set(data[cmptno], 0, x, (v));
			}
		}
		for (cmptno = 0; cmptno < numcmpts; ++cmptno) {
			if (image_writecmpt(image, cmptno, 0, y, hdr->width, 1,
			  data[cmptno])) {
				goto done;
			}
		}
	}

	ret = 0;

done:

	for (cmptno = 0; cmptno < numcmpts; ++cmptno) {
		if (data[cmptno]) {
			matrix_destroy(data[cmptno]);
		}
	}

	return ret;
}

/******************************************************************************\
* Miscellaneous functions.
\******************************************************************************/

static int pnm_getuint32(stream_t *in, uint_fast32_t *val)
{
	int_fast32_t v;
	if (pnm_getint32(in, &v)) {
		return -1;
	}
	if (v < 0) {
		return -1;
	}
	if (val) {
		*val = v;
	}
	return 0;
}

static int pnm_getint32(stream_t *in, int_fast32_t *val)
{
	int c;
	int s;
	int_fast32_t v;

	do {
		if ((c = pnm_getc(in)) == EOF) {
			return -1;
		}
	} while (isspace(c));

	s = 1;
	if (c == '-') {
		s = -1;
		if ((c = pnm_getc(in)) == EOF) {
			return -1;
		}
	}
	v = 0;
	while (isdigit(c)) {
		v = 10 * v + c - '0';
		if ((c = pnm_getc(in)) < 0) {
			return -1;
		}
	}
	if (!isspace(c)) {
		return -1;
	}
	*val = (s >= 0) ? v : (-v);

	return 0;
}

static int pnm_getc(stream_t *in)
{
	int c;
	for (;;) {
		if ((c = stream_getc(in)) == EOF) {
			return -1;
		}
		if (c != '#') {
			return c;
		}
		do {
			if ((c = stream_getc(in)) == EOF) {
				return -1;
			}
		} while (c != '\n' && c != '\r');
	}
}

static int pnm_getuint16(stream_t *in, uint_fast16_t *val)
{
	uint_fast16_t v;
	int c;

	if ((c = stream_getc(in)) == EOF) {
		return -1;
	}
	v = c & 0xff;
	if ((c = stream_getc(in)) == EOF) {
		return -1;
	}
	v = (v << 8) | (c & 0xff);
	*val = v;

	return 0;
}
