/*
 * Copyright (c) 1999-2000 Image Power, Inc. and the University of
 *   British Columbia.
 * Copyright (c) 2001 Michael David Adams.
 * All rights reserved.
 *
 * The complete licensing terms for this software can be found in the file
 * named "LICENSE" in this software distribution.
 */

/*
 * Portable Pixmap/Graymap Format Support
 *
 * $Id$
 */

/******************************************************************************\
* Includes.
\******************************************************************************/

#include <ctype.h>
#include <math.h>
#include <stdlib.h>
#include <assert.h>

#include "types.h"
#include "memalloc.h"
#include "tvparser.h"
#include "pnm.h"

/******************************************************************************\
* Local types.
\******************************************************************************/

typedef struct {
	bool bin;
} pnm_encopts_t;

typedef enum {
	OPT_TEXT
} optid_t;

taginfo_t opttab[] = {
	{OPT_TEXT, "text"},
	{-1, 0}
};

/******************************************************************************\
* Local function prototypes.
\******************************************************************************/

static int pnm_parseencopts(char *optstr, pnm_encopts_t *encopts);
static int pnm_puthdr(stream_t *out, pnm_hdr_t *hdr);
static int pnm_putuint16(stream_t *out, uint_fast16_t val);
static int pnm_putdata(stream_t *out, pnm_hdr_t *hdr, image_t *image);

/******************************************************************************\
* Save function.
\******************************************************************************/

int image_save_pnm(image_t *image, stream_t *out, char *optstr)
{
	uint_fast32_t width;
	uint_fast32_t height;
	uint_fast16_t cmptno;
	pnm_hdr_t hdr;
	pnm_encopts_t encopts;
	uint_fast16_t numcmpts;
	int prec;
	bool sgnd;

	/* Parse the encoder option string. */
	if (pnm_parseencopts(optstr, &encopts)) {
		fprintf(stderr, "invalid PNM encoder options specified\n");
		return -1;
	}

	numcmpts = image_numcmpts(image);
	width = image_cmptwidth(image, 0);
	height = image_cmptheight(image, 0);
	prec = image_cmptprec(image, 0);
	sgnd = image_cmptsgnd(image, 0);

	/* The PNM format is quite limited in the set of image geometries
	  that it can handle.  Here, we check to ensure that the image to
	  be saved can actually be represented reasonably accurately using the
	  PNM format. */
	/* All of the components must have the same width and height. */
	/* All of the components must have unsigned samples with the same
	  precision.*/
	/* All of the components must have their top-left corner located at
	  the origin. */
	for (cmptno = 0; cmptno < numcmpts; ++cmptno) {
		if (image_cmptwidth(image, cmptno) != width ||
		  image_cmptheight(image, cmptno) != height ||
		  image_cmptprec(image, cmptno) != prec ||
		  image_cmptsgnd(image, cmptno) != sgnd ||
		  image_cmpthstep(image, cmptno) != image_cmpthstep(image, 0) ||
		  image_cmptvstep(image, cmptno) != image_cmptvstep(image, 0) ||
		  image_cmpttlx(image, cmptno) != image_cmpttlx(image, 0) ||
		  image_cmpttly(image, cmptno) != image_cmpttly(image, 0)) {
			fprintf(stderr, "The PNM format cannot be used to represent an image with this geometry.\n");
			return -1;
		}
	}

	if (sgnd) {
		fprintf(stderr, "warning: support for signed sample data requires use of nonstandard extension to PNM format\n");
		fprintf(stderr, "You may not be able to read or correctly display the resulting PNM data with other software.\n");
	}

	/* The PNM format does not allow for binary data when the depth is
	  greater than eight. */
	if (prec > 8 || sgnd) {
		encopts.bin = false;
	}

	/* Initialize the header. */
	if (numcmpts == 1) {
		hdr.magic = encopts.bin ? PNM_MAGIC_BINPGM : PNM_MAGIC_TXTPGM;
	} else if (numcmpts == 3) {
		hdr.magic = encopts.bin ? PNM_MAGIC_BINPPM : PNM_MAGIC_TXTPPM;
	} else {
		return -1;
	}
	hdr.width = width;
	hdr.height = height;
	hdr.maxval = (1 << prec) - 1;

	/* Write the header. */
	if (pnm_puthdr(out, &hdr)) {
		return -1;
	}

	/* Write the image data. */
	if (pnm_putdata(out, &hdr, image)) {
		return -1;
	}

	/* Flush the output stream. */
	if (stream_flush(out)) {
		return -1;
	}

	return 0;
}

/******************************************************************************\
* Code for parsing options.
\******************************************************************************/

/* Parse the encoder options string. */
static int pnm_parseencopts(char *optstr, pnm_encopts_t *encopts)
{
	tvparser_t *tvp;
	int ret;

	tvp = 0;

	/* Initialize default values for encoder options. */
	encopts->bin = true;

	/* Create the tag-value parser. */
	if (!(tvp = tvparser_create(optstr ? optstr : ""))) {
		goto error;
	}

	/* Get tag-value pairs, and process as necessary. */
	while (!(ret = tvparser_next(tvp))) {
		switch (taginfo_nonull(taginfos_lookup(opttab,
		  tvparser_gettag(tvp)))->id) {
		case OPT_TEXT:
			encopts->bin = false;
			break;
		default:
			fprintf(stderr, "warning: ignoring invalid option %s\n",
			  tvparser_gettag(tvp));
			break;
		}	
	}
	if (ret < 0) {
		goto error;
	}

	/* Destroy the tag-value parser. */
	tvparser_destroy(tvp);

	return 0;

error:
	if (tvp) {
		tvparser_destroy(tvp);
	}
	return -1;
}

/******************************************************************************\
* Function for writing header.
\******************************************************************************/

/* Write the header. */
static int pnm_puthdr(stream_t *out, pnm_hdr_t *hdr)
{
	if (pnm_putuint16(out, hdr->magic)) {
		return -1;
	}
	stream_printf(out, "\n%lu %lu\n%lu\n", (unsigned long) hdr->width,
	  (unsigned long) hdr->height, (unsigned long) hdr->maxval);
	if (stream_error(out)) {
		return -1;
	}
	return 0;
}

/******************************************************************************\
* Functions for processing the sample data.
\******************************************************************************/

/* Write the image sample data. */
static int pnm_putdata(stream_t *out, pnm_hdr_t *hdr, image_t *image)
{
	int ret;
	uint_fast16_t cmptno;
	uint_fast32_t x;
	uint_fast32_t y;
	matrix_t *data[3];
	int numcmpts;
	int fmt;
	fix_t *d[3];
	fix_t v;
	int minval;
	int linelen;
	int n;
	char buf[256];

	ret = -1;
	fmt = pnm_fmt(hdr->magic);
	minval = -((int) hdr->maxval + 1);

	numcmpts = image_numcmpts(image);
	data[0] = 0;
	data[1] = 0;
	data[2] = 0;
	for (cmptno = 0; cmptno < numcmpts; ++cmptno) {
		if (!(data[cmptno] = matrix_create(1, hdr->width))) {
			goto done;
		}
	}

	for (y = 0; y < hdr->height; ++y) {
		for (cmptno = 0; cmptno < numcmpts; ++cmptno) {
			if (image_readcmpt(image, cmptno, 0, y, hdr->width, 1,
			  data[cmptno])) {
				goto done;
			}
			d[cmptno] = matrix_getref(data[cmptno], 0, 0);
		}
		linelen = 0;
		for (x = 0; x < hdr->width; ++x) {
			for (cmptno = 0; cmptno < numcmpts; ++cmptno) {
				v = *d[cmptno];
				if (v < minval) {
					v = minval;
				}
				if (v > ((int) hdr->maxval)) {
					v = hdr->maxval;
				}
				if (fmt == PNM_FMT_BIN) {
					stream_putc(out, (unsigned char)(v));
				} else {
					n = sprintf(buf, "%s%ld", ((!(!x && !cmptno)) ? " " : ""),
					  (long) v);
					if (linelen > 0 && linelen + n > PNM_MAXLINELEN) {
						stream_printf(out, "\n");
						linelen = 0;
					}
					stream_printf(out, "%s", buf);
					linelen += n;
				}
				++d[cmptno];
			}
		}
		if (fmt != PNM_FMT_BIN) {
			stream_printf(out, "\n");
			linelen = 0;
		}
		if (stream_error(out)) {
			goto done;
		}
	}

	ret = 0;

done:

	for (cmptno = 0; cmptno < numcmpts; ++cmptno) {
		if (data[cmptno]) {
			matrix_destroy(data[cmptno]);
		}
	}

	return ret;
}

/******************************************************************************\
* Miscellaneous functions.
\******************************************************************************/

/* Write a 16-bit unsigned integer to a stream. */
static int pnm_putuint16(stream_t *out, uint_fast16_t val)
{
	if (stream_putc(out, (unsigned char)(val >> 8)) == EOF ||
	  stream_putc(out, (unsigned char)(val & 0xff)) == EOF) {
		return -1;
	}
	return 0;
}
